import os
import re
import csv
from dataclasses import dataclass


def parse_conf(
    path: str = "./.conf",
) -> dict[str, str]:
    """Parses variables from .conf file

    Args:
        path (str): Path to file to parse. Defaults to ./.conf

    Returns:
        Dict[str, str]: Dictionary containing keys and value pairs of vars in file.
    """
    with open(path) as stream:
        contents = stream.read().strip()

    var_declarations = re.findall(r"^[a-zA-Z0-9_]+=.*$", contents, flags=re.MULTILINE)
    reader = csv.reader(var_declarations, delimiter="=")
    bash_vars = dict(reader)

    return bash_vars


def get_part_id() -> str:
    """Get part id from environment variable.
    Returns:
        str: The part id.
    """
    part_id = os.getenv("partId")
    if part_id is None:
        return ""
    return part_id


@dataclass
class Config:
    """Class that holds configuration values for the grader."""

    bind_mount: str = "/shared/submission/"
    submission_workdir: str = "./submission/"
    solution_workdir: str = "./solution/"
    solution_file: str = "solution.ipynb"
    solution_file_path: str = ""
    submission_file: str = "submission.ipynb"
    submission_file_path: str = ""
    part_id: str = get_part_id()
    latest_version: str = (
        parse_conf()["GRADER_VERSION"] if os.path.exists("./.conf") else ""
    )
    assignment_name: str = (
        parse_conf()["ASSIGNMENT_NAME"] if os.path.exists("./.conf") else ""
    )

    def __post_init__(self):
        # This is where we set the file paths after the instance variables are initialized
        self.solution_file_path = os.path.join(
            self.solution_workdir, self.solution_file
        )
        self.submission_file_path = os.path.join(
            self.submission_workdir, self.submission_file
        )



import ast
from types import ModuleType
from contextlib import nullcontext
from nbformat.notebooknode import NotebookNode
from .io import suppress_stdout_stderr


def compile_module(
    code_as_str: str,
    module_name: str,
    wipe_global_state: bool = False,
    verbose: bool = True,
) -> ModuleType:
    """Compiles the string representation of some code and returns a compiled module.
    Args:
        code_as_str (str): Code represented as string.
        module_name (str): Name of the module.
        wipe_global_state (bool): If true then no global state is compiled. Defaults to False.
        verbose (bool): Whether to print out streams as a result of compilation. Defaults to True.
    Returns:
        ModuleType: The actual module that can be used to call functions/variables/etc.
    """
    code_ast = ast.parse(code_as_str)

    if wipe_global_state:
        for node in code_ast.body[:]:
            if not isinstance(
                node, (ast.FunctionDef, ast.Import, ast.ImportFrom, ast.ClassDef)
            ):
                code_ast.body.remove(node)

    with nullcontext() if verbose else suppress_stdout_stderr():
        module = ModuleType(module_name)
        code = compile(code_ast, f"{module_name}.py", "exec")
        exec(code, module.__dict__)
        return module


def compile_partial_module(
    notebook: NotebookNode,
    module_name: str,
    verbose: bool = True,
    exit_on_error: bool = False,
    debug_mode: bool = False,
) -> ModuleType:
    """Iterates over the code cells of a notebook and includes the ones that run to the compiled module.
    Args:
        notebook (NotebookNode): Notebook from learner.
        module_name (str): Name of the module.
        verbose (bool): Whether to print out streams as a result of compilation. Defaults to True.
        exit_on_error (bool): Whether to stop compilation if an exception is found. Defaults to False.
        debug_mode (bool): Whether to print out cells where exceptions occurred. Defaults to False.
    Returns:
        ModuleType: The actual module that can be used to call functions/variables/etc.
    """
    code_cells = [cell.source for cell in notebook.cells if cell.cell_type == "code"]
    module = ModuleType(module_name)

    for i, cell_code in enumerate(code_cells):
        try:
            compiled_code = compile(cell_code, f"<cell {i}>", "exec")

            with nullcontext() if verbose else suppress_stdout_stderr():
                exec(compiled_code, module.__dict__)

        except Exception as e:
            if exit_on_error:
                if debug_mode:
                    print(
                        f"Error during execution of cell. Aborting full compilation.\n\nContents:\n\n{cell_code}\n\nException:\n\n{e}\n"
                    )
                break

            if debug_mode:
                print(
                    f"Error during execution of cell but kept going.\n\nContents:\n\n{cell_code}\n\nException:\n\n{e}\n"
                )
            continue

    return module



from collections.abc import Callable
from types import ModuleType

from .grading import LearnerSubmission, test_case

grading_function = Callable[..., list[test_case]]
learner_submission = ModuleType | LearnerSubmission
grading_wrapper = Callable[[learner_submission, ModuleType | None], grading_function]



import os
import argparse
from .io import (
    update_grader_and_notebook_version,
    update_notebook_version,
    tag_notebook,
    undeletable_notebook,
    uneditable_notebook,
    init_grader,
    generate_learner_version,
    grade_parts,
)
from .config import Config


def parse_dlai_grader_args() -> None:
    """Parses command line flags and performs desired actions"""
    parser = argparse.ArgumentParser(
        description="Helper library to build automatic graders for DLAI courses."
    )
    parser.add_argument(
        "-i",
        "--init",
        action="store_true",
        help="Initialize a grader workspace with common files and directories.",
    )
    parser.add_argument(
        "-u",
        "--upgrade",
        action="store_true",
        help="Upgrade the grader and notebook version.",
    )
    parser.add_argument(
        "-v",
        "--versioning",
        action="store_true",
        help="Add version to notebook metadata that matches current grader version.",
    )
    parser.add_argument(
        "-t",
        "--tag",
        action="store_true",
        help="Add graded tag to all code cells of notebook.",
    )
    parser.add_argument(
        "-ud",
        "--undeletable",
        action="store_true",
        help="Make all code cells of notebook not deletable.",
    )
    parser.add_argument(
        "-ue",
        "--uneditable",
        action="store_true",
        help="Make all non-graded code cells of notebook not editable.",
    )
    parser.add_argument(
        "-l",
        "--learner",
        action="store_true",
        help="Generate learner facing version.",
    )
    parser.add_argument(
        "-in",
        "--input_notebook",
        type=str,
        help="Path to input notebook to generate learner version from.",
    )
    parser.add_argument(
        "-out",
        "--output_notebook",
        type=str,
        help="Path to save the generated learner version to.",
    )
    parser.add_argument(
        "-g",
        "--grade",
        action="store_true",
        help="Grade using coursera_autograder tool.",
    )
    parser.add_argument(
        "-p",
        "--partids",
        type=str,
        help="Partids encoded as a single string separated by spaces.",
    )
    parser.add_argument(
        "-d",
        "--docker",
        type=str,
        help="Docker image to use for grading.",
    )
    parser.add_argument(
        "-m",
        "--memory",
        type=str,
        help="Memory to assign to the container.",
    )
    parser.add_argument(
        "-s",
        "--submission",
        type=str,
        help="Submission directory.",
    )
    args = parser.parse_args()
    c = Config()
    if args.init:
        init_grader()
    if args.upgrade:
        update_grader_and_notebook_version()
    if args.versioning:
        update_notebook_version(
            path="./mount/submission.ipynb", version=c.latest_version
        )
    if args.tag:
        tag_notebook("./mount/submission.ipynb")
    if args.undeletable:
        undeletable_notebook("./mount/submission.ipynb")
    if args.uneditable:
        uneditable_notebook("./mount/submission.ipynb")

    if args.learner:
        filename_source = "./mount/submission.ipynb"
        filename_target = f"./learner/{c.assignment_name}.ipynb"

        if not os.path.exists("./mount/") and (not args.input_notebook):
            print(
                "No mount/ directory found. Looking for .ipynb files in current directory.\n"
            )
            notebooks_current_directory = [
                f for f in os.listdir(".") if f.endswith(".ipynb")
            ]

            if len(notebooks_current_directory) > 1:
                print(
                    "More than one notebook found in current directory. Specify '--input_notebook' flag."
                )
                return

            filename_source = notebooks_current_directory[0]

        if args.input_notebook:
            filename_source = args.input_notebook

        filename_current_directory = filename_source.split("/")[-1]
        notebook_name_no_extension = filename_current_directory.split(".")[0]
        filename_target = f"./{notebook_name_no_extension}_learner.ipynb"

        if args.output_notebook:
            filename_target = args.output_notebook

        generate_learner_version(
            filename_source=filename_source,
            filename_target=filename_target,
        )
    if args.grade:
        if not args.partids:
            print("partids not provided")
            return
        if not args.docker:
            print("docker not provided")
            return
        if not args.memory:
            print("memory not provided")
            return
        if not args.submission:
            print("submission not provided")
            return

        grade_parts(args.partids, args.docker, args.submission, args.memory)



import json
import os
import shutil
import subprocess
import tarfile
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull
from pathlib import Path
from textwrap import dedent
from zipfile import ZipFile

import jupytext
import nbformat
from nbformat.notebooknode import NotebookNode

from .notebook import (
    add_metadata_all_code_cells,
    add_metadata_code_cells_without_pattern,
    solution_to_learner_format,
    tag_code_cells,
)
from .templates import load_templates


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull."""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def read_notebook(
    path: str,
) -> NotebookNode:
    """
    Reads a notebook found in the given path and returns a serialized version.

    Args:
        path (str): Path of the notebook file to read.

    Returns:
        NotebookNode: Representation of the notebook following nbformat convention.

    """
    return nbformat.read(path, as_version=nbformat.NO_CONVERT)


def tag_notebook(
    path: str,
) -> None:
    """
    Adds 'graded' tag to all code cells of a notebook.

    Args:
        path (str): Path to the notebook.

    """
    nb = read_notebook(path)
    nb = tag_code_cells(nb)
    jupytext.write(nb, path)


def undeletable_notebook(path: str) -> None:
    """
    Makes all code cells of a notebook non-deletable.

    Args:
        path (str): Path to the notebook.

    """
    nb = read_notebook(path)
    nb = add_metadata_all_code_cells(nb, {"deletable": False})
    jupytext.write(nb, path)


def uneditable_notebook(path: str) -> None:
    """
    Makes all non-graded code cells of a notebook non-editable.

    Args:
        path (str): Path to the notebook.

    """
    nb = read_notebook(path)
    nb = add_metadata_code_cells_without_pattern(
        nb, {"editable": False}, ignore_pattern="^# EDITABLE"
    )
    jupytext.write(nb, path)


def extract_tar(
    file_path: str,
    destination: str,
    post_cleanup: bool = True,
) -> None:
    """
    Extracts a tar file unto the desired destination.

    Args:
        file_path (str): Path to tar file.
        destination (str): Path where to save uncompressed files.
        post_cleanup (bool, optional): If true, deletes the compressed tar file. Defaults to True.

    """
    with tarfile.open(file_path, "r") as my_tar:
        my_tar.extractall(destination)

    if post_cleanup and os.path.exists(file_path):
        os.remove(file_path)


def extract_zip(
    file_path: str,
    destination: str,
    post_cleanup: bool = True,
) -> None:
    """Extracts a zip file unto the desired destination.

    Args:
        file_path (str): Path to zip file.
        destination (str): Path where to save uncompressed files.
        post_cleanup (bool, optional): If true, deletes the compressed zip file. Defaults to True.
    """
    with ZipFile(file_path, "r") as zip:
        zip.extractall(destination)

    if post_cleanup and os.path.exists(file_path):
        os.remove(file_path)


def send_feedback(
    score: float,
    msg: str,
    feedback_path: str = "/shared/feedback.json",
    err: bool = False,
) -> None:
    """
    Sends feedback to the learner.

    Args:
        score (float): Grading score to show on Coursera for the assignment.
        msg (str): Message providing additional feedback.
        feedback_path (str): Path where the json feedback will be saved. Defaults to /shared/feedback.json
        err (bool, optional): True if there was an error while grading. Defaults to False.

    """
    post = {"fractionalScore": score, "feedback": msg}
    print(json.dumps(post))

    with open(feedback_path, "w") as outfile:
        json.dump(post, outfile)

    if err:
        exit(1)

    exit(0)


def copy_submission_to_workdir(
    dir_origin: str = "/shared/submission/",
    dir_destination: str = "./submission/",
    file_name: str = "submission.ipynb",
) -> None:
    """Copies submission file from bind mount directory into working directory.
    Args:
        dir_origin (str): Origin directory.
        dir_destination (str): Target directory.
        file_name (str): Name of the file.
    """

    _from = os.path.join(dir_origin, file_name)
    _to = os.path.join(dir_destination, file_name)
    shutil.copyfile(_from, _to)


def update_grader_version() -> str:
    """Updates the grader version by 1 unit.

    Returns:
        str: New version of the grader.
    """
    with open("./.conf", "r") as f:
        lines = f.readlines()

    new_lines = []
    for l in lines:
        if ("GRADER_VERSION" in l) and ("TAG_ID" not in l):
            _, v = l.split("=")
            num_v = int(v)
            new_v = num_v + 1
            new_l = f"GRADER_VERSION={new_v}\n"
            new_lines.append(new_l)
            continue
        new_lines.append(l)

    with open("./.conf", "w") as f:
        f.writelines(new_lines)

    return str(new_v)


def update_notebook_version(
    path: str,
    version: str,
) -> None:
    """Updates notebook version to match the latest version of the grader.

    Args:
        path (str): Path to the notebook.
        version (str): Latest version of the grader to update the notebook to.
    """
    nb = read_notebook(path)
    metadata = nb.get("metadata")
    metadata.update({"grader_version": version})
    nb["metadata"] = metadata
    jupytext.write(nb, path)


def update_grader_and_notebook_version() -> None:
    """Updates the notebook and the grader at the same time."""
    latest_version = update_grader_version()
    update_notebook_version("./mount/submission.ipynb", latest_version)


def write_file_from_template(
    filename: str,
    template: str,
) -> None:
    """Writes the contents of a template unto a file.

    Args:
        filename (str): Name of the file.
        template (str): Template.
    """
    with open(filename, "w") as f:
        f.write(template)


def init_grader() -> None:
    """Initializes a grader directory."""

    template_dict = load_templates()
    write_file_from_template("./Dockerfile", template_dict["dockerfile"])
    write_file_from_template("./grader.py", template_dict["grader_py"])
    write_file_from_template("./Makefile", template_dict["makefile"])
    write_file_from_template("./.conf", template_dict["conf"])
    write_file_from_template("./entry.py", template_dict["entry_py"])
    # write_file_from_template(
    #     "./copy_assignment_to_submission.sh",
    #     template_dict["copy_assignment_to_submission_sh"],
    # )
    write_file_from_template("./requirements.txt", "dlai-grader")
    write_file_from_template("./.env", "")

    os.makedirs("learner")
    os.makedirs("mount")
    os.makedirs("submission")

    extra_file_name = template_dict["extra_file_name"]
    if extra_file_name:
        write_file_from_template(f"./learner/{extra_file_name}", "")
        write_file_from_template(f"./mount/{extra_file_name}", "")

    write_file_from_template(
        "./mount/submission.ipynb", template_dict["submission_ipynb"]
    )

    if "COPY data/ /grader/data/" in template_dict["dockerfile"]:
        os.makedirs("data")

    if "COPY solution/ /grader/solution/" in template_dict["dockerfile"]:
        os.makedirs("solution")


def generate_learner_version(
    filename_source: str,
    filename_target: str,
) -> None:
    """Generates the learning facing version of an assignment.

    Args:
        filename_source (str): Path to original notebook.
        filename_target (str): Path where to save reformatted notebook.
    """
    notebook = read_notebook(filename_source)

    for cell in notebook["cells"]:
        if cell["cell_type"] == "code":
            cell_code = cell["source"]
            fmt_code = solution_to_learner_format(cell_code)
            cell["source"] = fmt_code

    jupytext.write(notebook, filename_target)


def grade_parts(
    partids: str,
    image: str,
    sub_dir: str,
    mem: str,
) -> None:
    """Grades all parts using coursera_autograder tool.

    Args:
        partids (str): String encoding all part_ids separated by spaces.
        image (str): Name of the docker image to use.
        sub_dir (str): Directory including the submission.
        mem (str): Amount of memory to allocate for the container.
    """

    for p in partids.split(" "):
        print(f"\nGrading part_id: {p}\n")
        cmd = """coursera_autograder grade local $2 $3 '{"partId": "$1", "fileName": "submission.ipynb"}' --mem-limit $4"""

        cmd = cmd.replace("$1", f"{p}")
        cmd = cmd.replace("$2", image)
        cmd = cmd.replace("$3", sub_dir)
        cmd = cmd.replace("$4", mem)
        cmd = dedent(cmd)

        try:
            subprocess.run(cmd, shell=True, executable="/bin/bash")
        except Exception as e:
            print(f"There was an error grading with coursera_autograder. Details: {e}")


def generate_learner_file(
    filename_source: str,
    filename_target: str,
) -> None:
    """
    Generates the learning facing version of any file.

    Args:
        filename_source (str): Path to original notebook.
        filename_target (str): Path where to save reformatted notebook.

    """
    solution_code = Path(filename_source).read_text()

    # format the code to replace with placeholders
    fmt_code = solution_to_learner_format(solution_code)

    # save the learner files
    Path(filename_target).write_text(fmt_code, encoding="utf-8")



from types import ModuleType, FunctionType
from dlai_grader.grading import test_case, object_to_grade
from dlai_grader.types import grading_function, grading_wrapper, learner_submission


def part_1(
    learner_mod: learner_submission,
    solution_mod: ModuleType | None = None,
) -> grading_function:
    @object_to_grade(learner_mod, "learner_func")
    def g(learner_func: FunctionType) -> list[test_case]:
        cases: list[test_case] = []

        t = test_case()
        if not isinstance(learner_func, FunctionType):
            t.fail()
            t.msg = "learner_func has incorrect type"
            t.want = FunctionType
            t.got = type(learner_func)
            return [t]

        return cases

    return g


def handle_part_id(part_id: str) -> grading_wrapper:
    grader_dict: dict[str, grading_wrapper] = {
        "123": part_1,
    }
    return grader_dict[part_id]



import traceback
from pathlib import Path

from grader import handle_part_id

from dlai_grader.compiler import compile_partial_module
from dlai_grader.config import Config, get_part_id
from dlai_grader.grading import (
    LearnerSubmission,
    compute_grading_score,
    graded_obj_missing,
)
from dlai_grader.io import copy_submission_to_workdir, read_notebook, send_feedback
from dlai_grader.notebook import keep_tagged_cells


def notebook_grading(config, compile_solution=False):
    try:
        nb = read_notebook(config.submission_file_path)
    except Exception as e:
        msg = f"There was a problem reading your notebook. Details:\n{e!s}"
        send_feedback(0.0, msg, err=True)

    transformations = [keep_tagged_cells()]

    for t in transformations:
        nb = t(nb)

    try:
        learner_mod = compile_partial_module(nb, "learner_mod", verbose=False)
    except Exception as e:
        msg = f"There was a problem compiling the code from your notebook, please check that you saved before submitting. Details:\n{e!s}"
        send_feedback(0.0, msg, err=True)

    solution_mod = None
    if compile_solution:
        solution_nb = read_notebook(config.solution_file_path)

        for t in transformations:
            solution_nb = t(solution_nb)

        solution_mod = compile_partial_module(
            solution_nb,
            "solution_mod",
            verbose=False,
        )

    return learner_mod, solution_mod


def non_notebook_grading(config):
    if Path(config.submission_file_path).stat().st_size == 0:
        msg = "Dummy file detected. Make sure you saved the correct file before submitting."
        send_feedback(0.0, msg, err=True)

    try:
        with open(config.submission_file_path, "r") as file:
            contents = file.read()
    except Exception as e:
        msg = f"There was an error reading your submission. Details:\n{e!s}"
        send_feedback(0.0, msg, err=True)

    return LearnerSubmission(submission=contents)


def main() -> None:
    part_id = get_part_id()

    match part_id:
        case "456":
            try:
                copy_submission_to_workdir(file_name="{{EXTRA_FILE_NAME}}")
            except FileNotFoundError as e:
                msg = f"""File {{EXTRA_FILE_NAME}} required for grading not found.
                This can happen if the file was deleted or you submitted without editing it.
                Details:\n{e!s}"""
                send_feedback(0.0, msg, err=True)
            except Exception as e:
                msg = f"There was an issue handling your submission. Details:\n{e!s}"
            c = Config(submission_file="{{EXTRA_FILE_NAME}}")
            learner_mod = non_notebook_grading(c)
        case _:
            try:
                copy_submission_to_workdir()
            except FileNotFoundError as e:
                msg = f"""Notebook required for grading not found.
                This can happen if the file was deleted or you submitted without editing it.
                Details:\n{e!s}"""
                send_feedback(0.0, msg, err=True)
            except Exception as e:
                msg = f"There was an issue handling your submission. Details:\n{e!s}"
                send_feedback(0.0, msg, err=True)
            c = Config()
            learner_mod, _ = notebook_grading(c)

    g_func = handle_part_id(part_id)(learner_mod)

    try:
        cases = g_func()
    except Exception as e:
        msg = f"There was an error grading your submission. Details:\n{e!s}"
        send_feedback(0.0, msg, err=True)

    if graded_obj_missing(cases):
        msg = "Object required for grading not found. If you haven't completed the exercise this might be expected. Otherwise, check your solution as grader omits cells that throw errors."
        send_feedback(0.0, msg, err=True)

    score, feedback = compute_grading_score(cases)
    send_feedback(score, feedback)


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        msg = f"There was an error with the program. Exception:\n{e!s}.\nTraceback:\n{traceback.format_exc()}"
        send_feedback(0.0, msg, err=True)



import sys
import textwrap
from textwrap import dedent
import shutil
import os
from pathlib import Path


def generate_copy_assignment_script(
    extra_file_required="n",
    assignment_name="C1M2_Assignment.ipynb",
    extra_file_name="foo.txt",
):
    """
    Copy the appropriate copy_assignment_to_submission.sh script from templates depending on whether an extra file is required.

    Template files should be named:
        extrafile_n   (no extra file)
        extrafile_y   (with extra file)

    Returns:
        str: The final shell script contents after variable substitution.

    """
    if extra_file_required not in ("y", "n"):
        raise ValueError(f"Invalid extra_file_required value: {extra_file_required!r}")

    # Define template name pattern
    template_name = f"extrafile_{extra_file_required}"

    # Paths
    base_dir = Path(__file__).parent
    src = base_dir / "templates" / "copy_assignment_sh" / template_name
    dst = Path("copy_assignment_to_submission.sh")

    # Validate existence
    if not src.exists():
        raise FileNotFoundError(f"Template not found: {src}")

    # Read and substitute placeholders
    content = src.read_text(encoding="utf-8")
    content = content.replace("{{ASSIGNMENT_NAME}}", assignment_name).replace(
        "{{EXTRA_FILE_NAME}}", extra_file_name
    )

    # Write output
    dst.write_text(content, encoding="utf-8")
    dst.chmod(0o755)  # make executable
    return content


def copy_entry_script(
    sol_dir_required: str,
    non_notebook_grading: str,
    extra_file_name="foo.txt",
) -> str:
    # Validate inputs
    if sol_dir_required not in ("y", "n"):
        raise ValueError(f"Invalid sol_dir_required value: {sol_dir_required!r}")
    if non_notebook_grading not in ("y", "n"):
        raise ValueError(
            f"Invalid non_notebook_grading value: {non_notebook_grading!r}"
        )

    template_name = f"solution_{sol_dir_required}_file_{non_notebook_grading}.py"

    base_dir = Path(__file__).parent
    src = base_dir / "templates" / "entry_py" / template_name
    content = src.read_text(encoding="utf-8")
    content = content.replace("{{EXTRA_FILE_NAME}}", extra_file_name)
    dst = Path("entry.py")

    if not src.exists():
        raise FileNotFoundError(f"Template not found: {src}")

    # shutil.copy(src, dst)
    return content


def generate_dockerfile(data_dir_required="n", sol_dir_required="n"):
    """
    Generate a Dockerfile with optional data and solution directories.

    Args:
        data_dir_required (str): Include data directory if "y"
        sol_dir_required (str): Include solution directory if "y"

    """
    # Validate inputs
    if data_dir_required not in ("y", "n"):
        raise ValueError(f"Invalid data_dir_required value: {data_dir_required!r}")
    if sol_dir_required not in ("y", "n"):
        raise ValueError(f"Invalid sol_dir_required value: {sol_dir_required!r}")

    template_name = f"data_{data_dir_required}_solution_{sol_dir_required}"

    # Define paths
    base_dir = Path(__file__).parent
    src = base_dir / "templates" / "dockerfile" / template_name
    dst = Path("Dockerfile")

    # Ensure the source exists
    if not src.exists():
        raise FileNotFoundError(f"Template not found: {src}")

    # Copy template to current directory
    # shutil.copy(src, dst)

    # Return the Dockerfile contents
    return src.read_text(encoding="utf-8")


def copy_makefile() -> str:
    base_dir = Path(__file__).parent
    src = base_dir / "templates" / "Makefile"
    # content = src.read_text(encoding="utf-8")
    # content = content.replace("{{HARD_MEMORY}}", hard_memory)
    # content = content.replace("{{CPUS}}", cpus)
    # content = content.replace("{{SOFT_MEMORY}}", soft_memory)
    dst = Path("Makefile")

    if not src.exists():
        raise FileNotFoundError(f"Template not found: {src}")

    # shutil.copy(src, dst)
    return src.read_text(encoding="utf-8")


def copy_grader_py() -> str:
    base_dir = Path(__file__).parent
    src = base_dir / "templates" / "grader.py"
    dst = Path("grader.py")

    if not src.exists():
        raise FileNotFoundError(f"Template not found: {src}")

    # shutil.copy(src, dst)
    return src.read_text(encoding="utf-8")


def copy_submission_ipynb() -> str:
    base_dir = Path(__file__).parent
    src = base_dir / "templates" / "submission.ipynb"

    if not src.exists():
        raise FileNotFoundError(f"Template not found: {src}")

    # shutil.copy(src, dst)
    return src.read_text(encoding="utf-8")


def load_templates() -> dict[str, str]:
    specialization = input("Name of the specialization: ")
    course = input("Number of the course: ")
    module = input("Number of the module: ")

    grader_mvp = input(
        "Use minimum grader (no extra config)? y/n (leave empty for n): ",
    )

    grader_mvp = grader_mvp if grader_mvp else "n"
    if grader_mvp not in ["y", "n"]:
        print("invalid option selected")
        sys.exit(1)

    if grader_mvp == "n":
        unit_test_filename = input(
            "Filename for unit tests (leave empty for unittests): "
        )
        unit_test_filename = unit_test_filename if unit_test_filename else "unittests"
        # version = input("Version of the grader (leave empty for version 1): ")
        version = "1"

        data_dir_required = ""
        while data_dir_required not in ["y", "n"]:
            data_dir_required = input(
                "Do you require a data dir? y/n (leave empty for n): ",
            )
            if data_dir_required == "":
                data_dir_required = "n"
            # data_dir_required = data_dir_required if data_dir_required else "n"

        sol_dir_required = ""
        while sol_dir_required not in ["y", "n"]:
            sol_dir_required = input(
                "Do you require a solution file? y/n (leave empty for n): ",
            )
            if sol_dir_required == "":
                sol_dir_required = "n"

        non_notebook_grading = ""
        while non_notebook_grading not in ["y", "n"]:
            non_notebook_grading = input(
                "Will you grade a file different from a notebook? y/n (leave empty for n): ",
            )
            if non_notebook_grading == "":
                non_notebook_grading = "n"

        extra_file_name = ""
        if non_notebook_grading == "y":
            extra_file_name = input(
                "Name of the extra file to grade: ",
            )

        cpus = ""
        while cpus not in ["0.25", "0.5", "0.75", "1"]:
            cpus = input("CPU Units (leave empty for 0.25): ")
            if cpus == "":
                cpus = "0.25"

            if cpus not in ["0.25", "0.5", "0.75", "1"]:
                print(f"Options are: {['0.25', '0.5', '0.75', '1']}")

        soft_memory = ""
        soft_memory_options = [
            "512m",
            "768m",
            "1024m",
            "2048m",
            "4096m",
            "8192m",
            "1g",
            "2g",
            "4g",
            "8g",
        ]
        while soft_memory not in soft_memory_options:
            soft_memory = input("Memory Reservation (leave empty for 512m): ")
            if soft_memory == "":
                soft_memory = "512m"

            if soft_memory not in soft_memory_options:
                print(f"Options are: {soft_memory_options}")

        hard_memory = ""
        hard_memory_options = [
            "1024m",
            "2048m",
            "4096m",
            "8192m",
            "15000m",
            "1g",
            "2g",
            "4g",
            "8g",
            "15g",
        ]
        while hard_memory not in hard_memory_options:
            hard_memory = input("Memory Limit (leave empty for 1g): ")
            if hard_memory == "":
                hard_memory = "1g"

            if hard_memory not in hard_memory_options:
                print(f"Options are: {hard_memory_options}")

    if grader_mvp == "y":
        unit_test_filename = "unittests"
        version = "1"
        data_dir_required = "n"
        sol_dir_required = "n"
        non_notebook_grading = "n"
        extra_file_name = ""
        cpus = "0.25"
        soft_memory = "512m"
        hard_memory = "1g"

    dockerfile = generate_dockerfile(
        data_dir_required=data_dir_required,
        sol_dir_required=sol_dir_required,
    )

    conf = f"""
    ASSIGNMENT_NAME=C{course}M{module}_Assignment
    UNIT_TESTS_NAME={unit_test_filename}
    IMAGE_NAME={specialization}c{course}m{module}-grader
    GRADER_VERSION={version}
    TAG_ID=V$(GRADER_VERSION)
    SUB_DIR=mount
    MEMORY_LIMIT=4096
    CPUS={cpus}
    SOFT_MEMORY={soft_memory}
    HARD_MEMORY={hard_memory}
    """

    assignment_name = f"C{course}M{module}_Assignment.ipynb"

    copy_assignment_to_submission_sh = generate_copy_assignment_script(
        extra_file_required=non_notebook_grading,
        assignment_name=assignment_name,
        extra_file_name=extra_file_name,
    )

    makefile = copy_makefile()

    grader_py = copy_grader_py()

    submission_ipynb = copy_submission_ipynb()

    entry_py = copy_entry_script(
        sol_dir_required=sol_dir_required,
        non_notebook_grading=non_notebook_grading,
        extra_file_name=extra_file_name,
    )

    template_dict = {
        "dockerfile": dedent(dockerfile),
        "makefile": dedent(makefile),
        "conf": dedent(conf[1:]),
        "grader_py": dedent(grader_py),
        "entry_py": dedent(entry_py),
        "extra_file_name": extra_file_name,
        "submission_ipynb": submission_ipynb,
        # "copy_assignment_to_submission_sh": dedent(copy_assignment_to_submission_sh),
    }

    if extra_file_name:
        template_dict.update({"extra_file_name": extra_file_name})

    return template_dict



from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable
from types import ModuleType


@dataclass
class LearnerSubmission:
    """Class that represents a file from the learner. Useful when grading does not depend on the notebook."""

    submission: Any = None


learner_submission = ModuleType | LearnerSubmission


@dataclass
class test_case:
    """Class that represents a test case."""

    msg: str = ""
    want: Any = None
    got: Any = None
    failed: bool = False

    def fail(self):
        """Sets the failed attribute to True."""
        self.failed = True


@dataclass
class aggregated_test_case:
    """Class that represents an aggregated collection of test cases."""

    test_name: str
    tests: list[test_case]


def compute_grading_score(test_cases: list[test_case]) -> tuple[float, str]:
    """
    Computes the score based on the number of failed and total cases.

    Args:
        test_cases (List): Test cases.

    Returns:
        Tuple[float, str]: The grade and feedback message.

    """
    num_cases = len(test_cases)
    if num_cases == 0:
        msg = "The grader was unable to generate test cases for your implementation. This suggests a bug with your code, please revise your solution and try again."
        return 0.0, msg

    failed_cases = [t for t in test_cases if t.failed]
    score = 1.0 - len(failed_cases) / num_cases
    score = round(score, 2)

    if not failed_cases:
        msg = "All tests passed! ✅ Congratulations! 🎉"
        return 1.0, msg

    feedback_msg = ""
    # if print_passed_tests:
    #     passed_cases = [t for t in test_cases if not t.failed]
    #     for passed_case in passed_cases:
    #         feedback_msg += f"✅ {passed_case.msg}.\n"

    if failed_cases:
        for failed_case in failed_cases:
            feedback_msg += f"❌ {failed_case.msg}.\n"

            if failed_case.want:
                feedback_msg += f"Expected:\n{failed_case.want}\n"

            if failed_case.got:
                feedback_msg += f"but got:\n{failed_case.got}\n\n"

    return score, feedback_msg


def compute_aggregated_grading_score(
    aggregated_test_cases: list[aggregated_test_case],
) -> tuple[float, str]:
    """
    Computes the score based on the number of failed and total cases.

    Args:
        aggregated_test_cases (List): Aggregated test cases for every part.

    Returns:
        Tuple[float, str]: The grade and feedback message.

    """
    scores = []
    msgs = []

    for test_cases in aggregated_test_cases:
        feedback_msg = f"All tests passed for {test_cases.test_name}!\n"
        num_cases = len(test_cases.tests)
        failed_cases = [t for t in test_cases.tests if t.failed]
        score = 1.0 - len(failed_cases) / num_cases
        score = round(score, 2)
        scores.append(score)

        if failed_cases:
            feedback_msg = f"Details of failed tests for {test_cases.test_name}\n\n"
            for failed_case in failed_cases:
                feedback_msg += f"Failed test case: {failed_case.msg}.\nExpected:\n{failed_case.want},\nbut got:\n{failed_case.got}.\n\n"
        msgs.append(feedback_msg)

    final_score = sum(scores) / len(scores)
    final_score = round(final_score, 2)
    final_msg = "\n".join(msgs)

    return final_score, final_msg


def object_to_grade(
    origin_module: learner_submission,
    *attr_names: str,
) -> Callable[[Callable[[Any], list[test_case]]], Callable[[Any], list[test_case]]]:
    """
    Used as a parameterized decorator to get any number of attributes from a module.

    Args:
        origin_module (ModuleType): A module.
        *attrs_name (Tuple[str, ...]): Names of the attributes to extract from the module.

    """

    def middle(func):
        @wraps(func)
        def wrapper():
            original_attrs = []
            for atrr in attr_names:
                original_attr = getattr(origin_module, atrr, None)
                original_attrs.append(original_attr)

            return func(*original_attrs)

        return wrapper

    return middle


def print_feedback(test_cases: list[test_case]) -> None:
    """
    Prints feedback of public unit tests within notebook.

    Args:
        test_cases (List[test_case]): List of public test cases.

    """
    failed_cases = [t for t in test_cases if t.failed]
    feedback_msg = "\033[92m All tests passed!"

    if failed_cases:
        feedback_msg = ""
        for failed_case in failed_cases:
            feedback_msg += f"\033[91mFailed test case: {failed_case.msg}.\nExpected: {failed_case.want}\nGot: {failed_case.got}\n\n"

    print(feedback_msg)


def graded_obj_missing(test_cases: list[test_case]) -> bool:
    """
    Check if the object to grade was found in the learned module.

    Args:
        test_cases (List[test_case]): List of test cases.

    Returns:
        bool: True if object is missing. False otherwise.

    """
    if len(test_cases) == 1 and test_cases[0].got is type(None):
        return True

    return False



import re
from typing import Callable
import jupytext
from nbformat.notebooknode import NotebookNode
from .config import Config


def notebook_to_script(
    notebook: NotebookNode,
) -> str:
    """Converts a notebook into a python script serialized as a string.
    Args:
        notebook (NotebookNode): Notebook to convert into script.
    Returns:
        str: Python script representation as string.
    """
    return jupytext.writes(notebook, fmt="py:percent")


def cut_notebook(
    regex_pattern: str = "(grade)(.|[ \t]*)(up)(.|[ \t]*)(to)(.|[ \t]*)(here)",
) -> Callable[[NotebookNode], NotebookNode]:
    """Cuts a notebook, this allows for partial grading. Written as a closure so it can be consumed as a functional option for notebooks.
    Args:
        regex_pattern (str): Regexp pattern to look for. Cells after match will be omitted.
    Returns:
        Callable[[NotebookNode], NotebookNode]: A function that cuts the notebook.
    """

    def inner(
        notebook: NotebookNode,
    ) -> NotebookNode:
        """Cuts a notebook by excluding all cells after a pattern is matched.
        Args:
            notebook (NotebookNode): Notebook to filter.
        Returns:
            NotebookNode: The filtered notebook.
        """
        filtered_cells = []

        for cell in notebook["cells"]:
            filtered_cells.append(cell)

            if cell["cell_type"] == "code" and re.search(regex_pattern, cell["source"]):
                break

        notebook["cells"] = filtered_cells
        return notebook

    return inner


def keep_tagged_cells(
    tag: str = "graded",
) -> Callable[[NotebookNode], NotebookNode]:
    """Keeps tagged cells from a notebook. Written as a closure so it can be consumed as a functional option for notebooks.
    Args:
        tag (str): Tag to look for within cell's metadata. Defaults to "graded".
    Returns:
        Callable[[NotebookNode], NotebookNode]: A function that filters the notebook.
    """

    def inner(
        notebook: NotebookNode,
    ) -> NotebookNode:
        """Filters a notebook by including tagged cells.
        Args:
            notebook (NotebookNode): Notebook to filter.
        Returns:
            NotebookNode: The notebook with tagged cells.
        """
        filtered_cells = []

        for cell in notebook["cells"]:
            if ("tags" not in cell["metadata"]) or (
                tag not in cell["metadata"].get("tags")
            ):
                continue
            filtered_cells.append(cell)

        notebook["cells"] = filtered_cells
        return notebook

    return inner


def omit_tagged_cells(
    tag: str = "omit",
) -> Callable[[NotebookNode], NotebookNode]:
    """Omits tagged cells from a notebook. Written as a closure so it can be consumed as a functional option for notebooks.
    Args:
        tag (str): Tag to look for within cell's metadata. Defaults to "omit".
    Returns:
        Callable[[NotebookNode], NotebookNode]: A function that filters the notebook.
    """

    def inner(
        notebook: NotebookNode,
    ) -> NotebookNode:
        """Filters a notebook by excluding tagged cells.
        Args:
            notebook (NotebookNode): Notebook to filter.
        Returns:
            NotebookNode: The notebook without omitted cells.
        """
        filtered_cells = []

        for cell in notebook["cells"]:
            if ("tags" in cell["metadata"]) and (tag in cell["metadata"].get("tags")):
                continue
            filtered_cells.append(cell)

        notebook["cells"] = filtered_cells
        return notebook

    return inner


def partial_grading_enabled(
    notebook: NotebookNode,
    regex_pattern: str = "(grade)(.|[ \t]*)(up)(.|[ \t]*)(to)(.|[ \t]*)(here)",
) -> bool:
    """Checks for the existence of the partial grading tag.

    Args:
        notebook (NotebookNode): The notebook from the learner.
        regex_pattern (str, optional): Pattern to check. Defaults to "(grade)(.|[ \t]*)(up)(.|[ \t]*)(to)(.|[ \t]*)(here)".

    Returns:
        bool: True if the pattern if found, False otherwise.
    """
    for cell in notebook["cells"]:
        if cell["cell_type"] == "code" and re.search(regex_pattern, cell["source"]):
            return True
    return False


def get_named_cells(
    notebook: NotebookNode,
) -> dict:
    """Returns the named cells for cases when grading is done using cell's output.
    Args:
        notebook (NotebookNode): The notebook from the learner.
    Returns:
        dict: All named cells encoded as a dictionary.
    """
    named_cells = {}
    for cell in notebook["cells"]:
        metadata = cell["metadata"]
        if "name" not in metadata:
            continue
        named_cells.update({metadata.get("name"): cell})
    return named_cells


def tag_code_cells(
    notebook: NotebookNode,
    tag: str = "graded",
) -> NotebookNode:
    """Filters a notebook to exclude additional cells created by learners.
       Also used for partial grading if the tag has been provided.
    Args:
        notebook (NotebookNode): Notebook to filter.
        tag (str): The tag to include in the code cell's metadata. Defaults to "graded".
    Returns:
        NotebookNode: The filtered notebook.
    """
    filtered_cells = []

    for cell in notebook["cells"]:
        if cell["cell_type"] == "code":
            if "tags" not in cell["metadata"]:
                cell["metadata"]["tags"] = []

            tags = cell["metadata"]["tags"]

            if tag not in tags:
                tags.append(tag)
                cell["metadata"]["tags"] = tags

        filtered_cells.append(cell)

    notebook["cells"] = filtered_cells

    return notebook


def add_metadata_all_code_cells(
    notebook: NotebookNode,
    metadata: dict,
) -> NotebookNode:
    """Adds metadata to all code cells of a notebook.
    Args:
        notebook (NotebookNode): Notebook to filter.
        metadata (dict): The metadata which should be a key-value pair.
    Returns:
        NotebookNode: The notebook with the new metadata.
    """
    filtered_cells = []

    for cell in notebook["cells"]:
        if cell["cell_type"] == "code":
            current_metadata = cell["metadata"]
            current_metadata.update(metadata)
            cell["metadata"] = current_metadata

        filtered_cells.append(cell)

    notebook["cells"] = filtered_cells

    return notebook


def add_metadata_code_cells_with_pattern(
    notebook: NotebookNode, metadata: dict, regex_pattern: str = "w[1-9]_unittest"
) -> NotebookNode:
    """Adds metadata to code cells of a notebook that match a regexp pattern.
    Args:
        notebook (NotebookNode): Notebook to filter.
        metadata (dict): The metadata which should be a key-value pair.
        regex_pattern (str, optional): Pattern to check. Defaults to "w[1-9]_unittest".
    Returns:
        NotebookNode: The notebook with the new metadata.
    """
    filtered_cells = []

    for cell in notebook["cells"]:
        if cell["cell_type"] == "code" and re.search(regex_pattern, cell["source"]):
            current_metadata = cell["metadata"]
            current_metadata.update(metadata)
            cell["metadata"] = current_metadata

        filtered_cells.append(cell)

    notebook["cells"] = filtered_cells

    return notebook


def add_metadata_code_cells_without_pattern(
    notebook: NotebookNode,
    metadata: dict,
    match_pattern: str = "^# GRADED",
    ignore_pattern: str = None,
) -> NotebookNode:
    """Adds metadata to code cells of a notebook that don't match a regexp pattern and aren't ignored by another pattern.

    Args:
        notebook (NotebookNode): Notebook to filter.
        metadata (dict): The metadata which should be a key-value pair.
        match_pattern (str, optional): Pattern to check which cells to add metadata to. Defaults to "^# GRADED ".
        ignore_pattern (str, optional): Pattern to check which cells to ignore. If a cell matches this pattern, it will be skipped.

    Returns:
        NotebookNode: The notebook with the new metadata.
    """
    for cell in notebook["cells"]:
        if cell["cell_type"] == "code":
            if ignore_pattern and re.search(ignore_pattern, cell["source"]):
                continue

            if not re.search(match_pattern, cell["source"]):
                current_metadata = cell.get("metadata", {})
                current_metadata.update(metadata)
                cell["metadata"] = current_metadata

    return notebook


def notebook_version(
    notebook: NotebookNode,
) -> str:
    """Returns dlai version of a notebook.

    Args:
        notebook (NotebookNode): A notebook.

    Returns:
        str: Version encoded as string.
    """
    return notebook.get("metadata").get("grader_version")


def notebook_is_up_to_date(
    notebook: NotebookNode,
) -> bool:
    """Determines if a notebook is up-to-date with latest grader version.

    Args:
        notebook (NotebookNode): A notebook.

    Returns:
        bool: True if both versions match, False otherwise.
    """
    version = notebook_version(notebook)
    c = Config()
    if version != c.latest_version:
        return False
    return True


def get_indentation(
    line: str,
) -> str:
    """Gets the indentation of a string.

    Args:
        line (str): A line of code.

    Returns:
        str: The indentation of that line.
    """
    spaces = []

    for char in line:
        if not char.isspace():
            break
        spaces.append(char)

    return "".join(spaces)


def solution_to_learner_format(
    cell_code: str,
    tag_start: str = "START CODE HERE",
    tag_end: str = "END CODE HERE",
) -> str:
    """Reformats the code in a cell by applying the rules defined by tags.

    Args:
        cell_code (str): The code of a notebook's cell represented as a string.
        tag_start (str, optional): Tag of start of code section. Defaults to "START CODE HERE".
        tag_end (str, optional): Tag of end of code section. Defaults to "END CODE HERE".

    Returns:
        str: Reformatted code.
    """

    tag_start_omit_block = "START OMIT BLOCK"
    tag_end_omit_block = "END OMIT BLOCK"
    tag_omit = "@OMIT"
    tag_keep = "@KEEP"
    tag_replace_equals = "@REPLACE EQUALS"
    tag_replace = "@REPLACE"

    code_block = False
    omit_block = False

    formatted_lines = []

    for ln in cell_code.splitlines():
        if tag_start in ln:
            code_block = True

        if tag_end in ln:
            code_block = False

        if tag_start_omit_block in ln:
            omit_block = True
            continue

        if tag_end_omit_block in ln:
            omit_block = False
            continue

        if (tag_omit in ln) or omit_block:
            continue

        code = ln.strip()
        first_cmd = code.split(" ")[0]

        if (tag_replace not in ln) and (
            first_cmd == "if" or first_cmd == "for" or first_cmd == "while"
        ):
            formatted_lines.append(ln)
            continue

        if tag_keep in ln:
            splitted = ln.split(f"# {tag_keep}")
            new_ln = splitted[0]
            formatted_lines.append(new_ln)
            continue

        if (tag_start not in ln) and code_block:
            if tag_replace_equals in ln:
                splitted = ln.split(tag_replace_equals)
                new_val = splitted[1]
                splitted_eq = splitted[0].split("=")
                new_ln = splitted_eq[0] + "=" + new_val
                formatted_lines.append(new_ln)
                continue

            if tag_replace in ln:
                ind = get_indentation(ln)
                splitted = ln.split(tag_replace)
                new_ln = ind + splitted[1].strip()
                formatted_lines.append(new_ln)
                continue

            if (" = " in ln) or (" += " in ln):
                splitted = ln.split("=")
                new_ln = splitted[0] + "= None"
                formatted_lines.append(new_ln)
                continue

        formatted_lines.append(ln)

    return "\n".join(formatted_lines)
